
from mindsdb_sdk.__about__ import __package_name__ as name, __version__

from .server import Server


def connect(url: str = None, email: str = None, password: str = None) -> Server:
    """
    Create connection to mindsdb server

    :param url: url to mindsdb server
    :param email: user email to login (for cloud version)
    :param password: user password to login (for cloud version)
    :return: Server object

    Examples
    --------

    >>> import mindsdb_sdk

    Connect to local server

    >>> server = mindsdb_sdk.connect()
    >>> server = mindsdb_sdk.connect('http://127.0.0.1:47334')

    Connect to cloud server

    >>> server = mindsdb_sdk.connect(email='a@b.com', password='-')
    >>> server = mindsdb_sdk.connect('https://cloud.mindsdb.com', email='a@b.com', password='-')

    """

    if url is None:
        if email is not None:
            # default is cloud
            url = 'https://cloud.mindsdb.com'
        else:
            # is local
            url = 'http://127.0.0.1:47334'

    return Server(url=url, email=email, password=password)



from typing import Union, List

import pandas as pd

from mindsdb_sql.parser.dialects.mindsdb import CreatePredictor, CreateView, DropPredictor
from mindsdb_sql.parser.ast import DropView, Identifier, Delete, Star, Select

from mindsdb_sdk.utils import dict_to_binary_op
from mindsdb_sdk.model import Model, ModelVersion
from mindsdb_sdk.query import Query, View


class Project:
    """
    Allows to work with project: to manage models and views inside of it or call raw queries inside of project

    Queries
    ----------
    Making prediciton using sql:

    >>> query = project.query('select * from database.table join model1')
    >>> df = query.fetch()

    Making time series prediction:

    >>> df = project.query('''
    ...      SELECT m.saledate as date, m.ma as forecast
    ...     FROM mindsdb.house_sales_model as m
    ...     JOIN example_db.demo_data.house_sales as t
    ...     WHERE t.saledate > LATEST AND t.type = 'house'
    ...     AND t.bedrooms=2
    ...     LIMIT 4;
    ...    ''').fetch()

    Views
    ----------
    Get:

    >>> views = project.list_views()
    >>> view = views[0]

    By name:

    >>> view project.get_view('view1')

    Create:

    >>> view = project.create_view(
    ...   'view1',
    ...   database='example_db',  # optional, can also be database object
    ...   query='select * from table1'
    ...)

    Create using query object:

    >>> view = project.create_view(
    ...   'view1',
    ...   query=database.query('select * from table1')
    ...)

    Getting data:

    >>> view.filter(a=1, b=2)
    >>> view.limit(100)
    >>> df = view.fetch()

    Drop view:

    >>> project.drop_view('view1')


    Models
    ----------

    Get:

    >>> models = project.list_models()
    >>> model = models[0]

    Get version:

    >>> models = project.list_models(with_versions=True)
    >>> model = models[0]

    By name:

    >>> model = project.get_model('model1')
    >>> model = project.get_model('model1', version=2)

    Versions

    List model versions

    >>> models = model.list_versions()
    >>> model = models[0]  # Model object


    Get info

    >>> print(model.status)
    >>> print(model.data)

    Update model data from server

    >>> model.refresh()

    Create

    Create, using params and qeury as string

    >>> model = project.create_model(
    ...   'rentals_model',
    ...   predict='price',
    ...   engine='lightwood',
    ...   database='example_db',
    ...   query='select * from table',
    ...   options={
    ...       'module': 'LightGBM'
    ...   },
    ...   timeseries_options={
    ...       'order': 'date',
    ...       'group': ['a', 'b']
    ...   }
    ...)

    Create, using deferred query. 'query' will be executed and converted to dataframe on mindsdb backend.

    >>> query = database.query('select * from table')
    >>> model = project.create_model(
    ...   'rentals_model',
    ...   predict='price',
    ...   query=query,
    ...)

    Usng model

    Dataframe on input

    >>> result_df = model.predict(df_rental)
    >>> result_df = model.predict(df_rental, params={'a': 'q'})

    Deferred query on input

    >>> result_df = model.predict(query, params={'': ''})

    Time series prediction

    >>> query = database.query('select * from table1 where type="house" and saledate>latest')
    >>> model.predict(query)

    The join model with table in raw query

    >>> result_df = project.query('''
    ...  SELECT m.saledate as date, m.ma as forecast
    ...   FROM mindsdb.house_sales_model as m
    ...   JOIN example_db.demo_data.house_sales as t
    ...  WHERE t.saledate > LATEST AND t.type = 'house'
    ...   AND t.bedrooms=2
    ...  LIMIT 4;
    ...''').fetch()


    Model managing

    Adjusting

    >>> model.adjust(query)
    >>> model.adjust('select * from demo_data.house_sales', database='example_db')
    >>> model.adjust(query, params={'x': 2})

    Retraining

    >>> model.retrain(query)
    >>> model.retrain('select * from demo_data.house_sales', database='example_db')
    >>> model.retrain(query, params={'x': 2})

    Describe

    >>> df_info = model.describe()
    >>> df_info = model.describe('features')

    Change active version

    >>> model.set_active(version=3)

    Drop

    >>> project.drop_model('rentals_model')
    >>> project.drop_model_version('rentals_model', version=10)

    """

    def __init__(self, server, name):
        self.name = name
        self.server = server
        self.api = server.api

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name})'

    def query(self, sql: str) -> Query:
        """
        Execute raw query inside of project

        :param sql: sql query
        :return: Query object
        """
        return Query(self.api, sql, database=self.name)

    def _list_views(self):
        df = self.api.objects_tree(self.name)
        df = df[df.type == 'view']

        return list(df['name'])

    def list_views(self) -> List[View]:
        """
        Show list of views in project

        :return: list of View objects
        """
        return [View(self, name) for name in self._list_views()]

    def create_view(self, name: str, sql: Union[str, Query], database: str = None) -> View:
        """
        Create new view in project and return it

        :param name: name of the view
        :param sql: sql query as string or query object
        :param database: datasource of the view (where input sql will be executed)
        :return: View object
        """
        if isinstance(sql, Query):
            database = sql.database
            sql = sql.sql
        elif not isinstance(sql, str):
            raise ValueError()

        if database is not None:
            database = Identifier(database)
        ast_query = CreateView(
            name=Identifier(name),
            query_str=sql,
            from_table=database
        )

        self.query(ast_query.to_string()).fetch()
        return View(self, name)

    def drop_view(self, name: str):
        """
        Drop view from project

        :param name: nave of the view
        """

        ast_query = DropView(names=[name])

        self.query(ast_query.to_string()).fetch()

    def get_view(self, name: str) -> View:
        """
        Get view by name from project

        :param name: name of the view
        :return: View object
        """

        if name not in self._list_views():
            raise AttributeError("View doesn't exist")
        return View(self, name)

    def list_models(self, with_versions: bool = False,
                    name: str = None,
                    version: int = None) -> List[Union[Model, ModelVersion]]:
        """
        List models (or model versions) in project

        If with_versions = True it shows all models with version (executes 'select * from models_versions')
        Otherwise it shows only models (executes 'select * from models')

        :param with_versions: show model versions
        :param name: to show models or versions only with selected name, optional
        :param version: to show model or versions only with selected version, optional
        :return: list of Model or ModelVersion objects
        """

        table = 'models'
        model_class = Model
        if with_versions:
            table = 'models_versions'
            model_class = ModelVersion

        filters = {}
        if name is not None:
            filters['NAME'] = name
        if version is not None:
            filters['VERSION'] = version

        ast_query = Select(
            targets=[Star()],
            from_table=Identifier(table),
            where=dict_to_binary_op(filters)
        )
        df = self.query(ast_query.to_string()).fetch()

        # columns to lower case
        cols_map = {i: i.lower() for i in df.columns}
        df = df.rename(columns=cols_map)

        return [
            model_class(self, item)
            for item in df.to_dict('records')
        ]

    def create_model(self, name: str, predict: str, engine: str = None,
                     query: Union[str, Query] = None, database: str = None,
                     options: dict = None, timeseries_options: dict = None) -> Model:
        """
        Create new model in project and return it

        If query/database is passed, it will be executed on mindsdb side

        :param name: name of the model
        :param predict: prediction target
        :param engine: ml engine for new model, default is mindsdb
        :param query: sql string or Query object to get data for training of model, optional
        :param database: database to get data for training, optional
        :param options: parameters for model, optional
        :param timeseries_options: parameters for forecasting model
        :return: created Model object, it can be still in training state
        """
        if isinstance(query, Query):
            database = query.database
            query = query.sql
        elif isinstance(query, pd.DataFrame):
            raise NotImplementedError('Dataframe as input for training model is not supported yet')

        if database is not None:
            database = Identifier(database)

        ast_query = CreatePredictor(
            name=Identifier(name),
            query_str=query,
            integration_name=database,
            targets=[Identifier(predict)],
        )

        if timeseries_options is not None:
            if 'group' in timeseries_options:
                group = timeseries_options['group']
                if not isinstance(group, list):
                    group = [group]
                ast_query.group_by = [Identifier(i) for i in group]
            if 'order' in timeseries_options:
                ast_query.order_by = [Identifier(timeseries_options['order'])]
            if 'window' in timeseries_options:
                ast_query.window = timeseries_options['window']
            if 'horizon' in timeseries_options:
                ast_query.horizon = timeseries_options['horizon']
        if options is None:
            options = {}
        if engine is not None:
            options['engine'] = engine

        df = self.query(ast_query.to_string()).fetch()
        if len(df) > 0:
            data = dict(df.iloc[0])
            # to lowercase
            data = {k.lower(): v for k,v in data.items()}

            return Model(self, data)

    def get_model(self, name: str, version: int = None) -> Union[Model, ModelVersion]:
        """
         Get model by name from project

         if version is passed it returns ModelVersion object with specific version

        :param name: name of the model
        :param version: version of model, optional
        :return: Model or ModelVersion object
        """
        if version is not None:
            ret = self.list_models(with_versions=True, name=name, version=version)
        else:
            ret = self.list_models(name=name)
        if len(ret) == 0:
            raise AttributeError("Model doesn't exist")
        elif len(ret) == 1:
            return ret[0]
        else:
            raise RuntimeError('Several models with the same name/version')

    def drop_model(self, name: str):
        """
        Drop model from project with all versions

        :param name: name of the model
        """
        ast_query = DropPredictor(name=name)
        self.query(ast_query.to_string()).fetch()

    def drop_model_version(self, name: str, version: int):
        """
        Drop version of the model

        :param name: name of the model
        :param version: version to drop
        """
        ast_query = Delete(
            table=Identifier('models_versions'),
            where=dict_to_binary_op({
                'name': name,
                'version': version
            })
        )
        self.query(ast_query.to_string()).fetch()




from typing import List

from mindsdb_sql.parser.ast import Identifier

from mindsdb_sdk.query import Query, Table


class Database:
    """
    Allows to work with database (datasource): to use tables and make raw queries

    To run native query
    At this moment query is just saved in Qeury object and not executed

    >>> query = database.query('select * from table1') # returns Query

    This command sends request to server to execute query and return dataframe

    >>> df = query.fetch()

    Wortking with tables:
    Get table as Query object

    >>> table = database.get_table('table1')

    Filter and limit

    >>> table.filter(a=1, b='2')
    >>> table.limit(3)

    Get content of table as dataframe. At that moment query will be sent on server and executed

    >>> df = table.fetch()

    Creating table

    From query:

    >>> table = database.create_table('table2', query)

    From other table

    >>> table2 = database.create_table('table2', table)

    """

    def __init__(self, server, name):
        self.server = server
        self.name = name
        self.api = server.api

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name})'

    def query(self, sql: str) -> Query:
        """
        Make raw query to integration

        :param sql: sql of the query
        :return: Query object
        """
        return Query(self.api, sql, database=self.name)

    def _list_tables(self):
        df = self.query('show tables').fetch()

        # first column
        return list(df[df.columns[0]])

    def list_tables(self) -> List[Table]:
        """
        Show list of tables in integration

        :return: list of Table objects
        """
        return [Table(self, name) for name in self._list_tables()]

    def get_table(self, name: str) -> Table:
        """
        Get table by name

        :param name: name of table
        :return: Table object
        """

        if name not in self._list_tables():
            if '.' not in name:
                # fixme: schemas not visible in 'show tables'
                raise AttributeError("Table doesn't exist")
        return Table(self, name)

    def create_table(self, name: str, query: Query, replace: bool = False) -> Table:
        """
        Create new table and return it.

        On mindsdb server it executes command:
        `insert into a (select ...)`

        or if replace is True
        `create table a (select ...)`

        'select ...' is extracted from input Query

        :param name: name of table
        :param query: Query object
        :param replace: if true,
        :return: Table object
        """

        if not isinstance(query, Query):
            raise NotImplementedError

        # # query can be in different database: wrap to NativeQuery
        # ast_query = CreateTable(
        #     name=Identifier(name),
        #     is_replace=is_replace,
        #     from_select=Select(
        #         targets=[Star()],
        #         from_table=NativeQuery(
        #             integration=Identifier(data.database),
        #             query=data.sql
        #         )
        #     )
        # )
        # self.query(ast_query.to_string()).fetch()

        # call in query database
        table = Identifier(parts=[self.name, name])

        replace_str = ''
        if replace:
            replace_str = ' or replace'

        self.api.sql_query(
            f'create{replace_str} table {table.to_string()} ({query.sql})',
            database=query.database
        )

        return Table(self, name)



from __future__ import annotations

from typing import List, Union

import pandas as pd

from mindsdb_sql.parser.dialects.mindsdb import RetrainPredictor, AdjustPredictor
from mindsdb_sql.parser.ast import Identifier, Select, Star, Join, Update, Describe, Constant
from mindsdb_sql import parse_sql
from mindsdb_sql.planner.utils import query_traversal

from mindsdb_sdk.utils import dict_to_binary_op
from mindsdb_sdk.query import Query


class Model:
    def __init__(self, project, data):
        self.project = project

        self.data = data
        self.name = data['name']
        self.version = None

    def __repr__(self):
        version = ''
        if self.version is not None:
            version = f', version={self.version}'
        return f'{self.__class__.__name__}({self.name}{version}, status={self.data["status"]})'

    def _get_identifier(self):
        parts = [self.project.name, self.name]
        if self.version is not None:
            parts.append(str(self.version))
        return Identifier(parts=parts)

    def predict(self, data: Union[pd.DataFrame, Query], params: dict = None) -> pd.DataFrame:
        """
        Make prediction using model

        if data is dataframe it uses /model/predict http method and sends dataframe over it
        if data is select query with one table it replaces table to jon table and predictor
        and sends query over sql/query http method

        if data is select from join other complex query it modifies query to:
          'select from (input query) join model' and sends it over sql/query http method

        :param data: dataframe or Query object as input to predictor
        :param params: parameters for predictor, optional
        :return: dataframe with result of prediction
        """
        if isinstance(data, Query):
            # create join from select if it is simple select
            ast_query = parse_sql(data.sql, dialect='mindsdb')
            if isinstance(ast_query, Select) and isinstance(ast_query.from_table, Identifier):
                # inject aliases
                if ast_query.from_table.alias is None:
                    alias = 't'
                    ast_query.from_table.alias = Identifier(alias)
                else:
                    alias = ast_query.from_table.alias.parts[-1]

                def inject_alias(node, is_table, **kwargs):
                    if not is_table:
                        if isinstance(node, Identifier):
                            if node.parts[0] != alias:
                                node.parts.insert(0, alias)

                query_traversal(ast_query, inject_alias)

                # replace table with join
                ast_query.from_table = Join(
                    join_type='join',
                    left=ast_query.from_table,
                    right=self._get_identifier()
                )
            else:
                # wrap query to subselect
                ast_query.parentheses = True
                ast_query = Select(
                    targets=[Star()],
                    from_table=Join(
                        join_type='join',
                        left=ast_query,
                        right=self._get_identifier()
                    )
                )
            if params is not None:
                ast_query.using = params
            # execute in query's database
            return self.project.api.sql_query(ast_query.to_string(), database=data.database)

        elif isinstance(data, pd.DataFrame):
            return self.project.api.model_predict(self.project.name, self.name, data,
                                                  params=params, version=self.version)
        else:
            raise ValueError('Unknown input')

    def get_status(self) -> str:
        """
        Refresh model data and return status of model

        :return: model status
        """
        self.refresh()
        return self.data['status']

    def refresh(self):
        """
        Refresh model data from mindsdb server
        Model data can be changed during training process

        :return: model data
        """
        model = self.project.get_model(self.name, self.version)
        self.data = model.data
        return self.data

    def adjust(self,
               query: Union[str, Query] = None,
               database: str = None,
               options: dict = None,
               engine: str = None) -> Union[Model, ModelVersion]:
        """
        Call adjust of the model

        :param query: sql string or Query object to get data for adjusting, optional
        :param database: database to get data for adjusting, optional
        :param options: parameters for adjusting model, optional
        :param engine: ml engine, optional
        :return: Model object
        """
        return self._retrain(ast_class=AdjustPredictor,
                             query=query, database=database,
                             options=options, engine=engine)

    def retrain(self,
               query: Union[str, Query] = None,
               database: str = None,
               options: dict = None,
               engine: str = None) -> Union[Model, ModelVersion]:
        """
        Call retrain of the model

        :param query: sql string or Query object to get data for retraining, optional
        :param database: database to get data for retraining, optional
        :param options: parameters for retraining model, optional
        :param engine: ml engine, optional
        :return: Model object
        """
        return self._retrain(ast_class=RetrainPredictor,
                             query=query, database=database,
                             options=options, engine=engine)

    def _retrain(self,
                 ast_class,
                 query: Union[str, Query] = None,
                 database:str = None,
                 options:dict = None,
                 engine:str = None):
        if isinstance(query, Query):
            database = query.database
            query = query.sql
        elif isinstance(query, pd.DataFrame):
            raise NotImplementedError('Dataframe as input for training model is not supported yet')

        if database is not None:
            database = Identifier(database)

        if options is None:
            options = {}
        if engine is not None:
            options['engine'] = engine

        ast_query = ast_class(
            name=self._get_identifier(),
            query_str=query,
            integration_name=database,
            using=options or None,
        )

        data = self.project.query(ast_query.to_string()).fetch()
        data = {k.lower(): v for k, v in data.items()}

        # return new instance
        base_class = self.__class__
        return base_class(self.project, data)

    def describe(self, type: str = None) -> pd.DataFrame:
        """
        Return description of the model

        :param type: describe type (for lightwood is models, ensemble, features), optional
        :return: dataframe with result of description
        """
        if self.version is not None:
            raise NotImplementedError

        identifier = self._get_identifier()
        if type is not None:
            identifier.parts.append(type)
        ast_query = Describe(identifier)
        return self.project.query(ast_query.to_string()).fetch()

    def list_versions(self) -> List[ModelVersion]:
        """
        Show list of model versions

        :return: list ModelVersion objects
        """
        return self.project.list_models(with_versions=True, name=self.name)

    def get_version(self, num: int) -> ModelVersion:
        """
        Get model version by number

        :param num: version number
        :return: ModelVersion object
        """

        num = int(num)
        for m in self.project.list_models(with_versions=True, name=self.name):
            if m.version == num:
                return m
        raise ValueError('Version is not found')

    def set_active(self, version: int):
        """
        Change model active version

        :param version: version to set active
        """
        ast_query = Update(
            table=Identifier('models_versions'),
            update_columns={
                'active': Constant(1)
            },
            where=dict_to_binary_op({
                'name': self.name,
                'version': version
            })
        )
        self.project.query(ast_query.to_string()).fetch()
        self.refresh()


class ModelVersion(Model):
    def __init__(self, project, data):

        super().__init__(project, data)

        self.version = data['version']



from typing import List

from mindsdb_sql.parser.dialects.mindsdb import CreateDatabase
from mindsdb_sql.parser.ast import DropDatabase

from mindsdb_sdk.connectors.rest_api import RestAPI
from .database import Database
from .project import Project


class Server:
    """
    Server instance allows to manipulate project and databases (integration) on mindsdb server
    Example if usage:

    Databases
    ----------

    >>> databases = server.list_databases()
    >>> database = databases[0] # Database type object

    # create

    >>> database = server.create_database('example_db',
    ...                                 type='postgres',
    ...                                 connection_args={'host': ''})

    # drop database

    >>> server.drop_database('example_db')

    # get existing

    >>> database = server.get_database('example_db')

    Projects
    ----------

    # list of projects

    >>> projects = server.list_projects()
    >>> project = projects[0]  # Project type object


    # create

    >>> project = server.create_project('proj')

    # drop

    >>> server.drop_project('proj')

    # get existing

    >>> project = server.get_project('proj')

    >>> project = server.get_project()  # default is mindsdb project

    """

    def __init__(self, url: str = None, email: str = None, password: str = None):
        self.api = RestAPI(url, email, password)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.api.url})'

    def _list_databases(self):
        data = self.api.sql_query(
            "select NAME from information_schema.databases where TYPE='data'"
        )
        return list(data.NAME)

    def list_databases(self) -> List[Database]:
        """
        Show list of integrations (databases) on server

        :return: list of Database objects
        """
        return [Database(self, name) for name in self._list_databases()]

    def create_database(self, name: str, engine: str, connection_args: dict) -> Database:
        """
        Create new integration and return it

        :param name: Identifier for the integration to be created
        :param engine: Engine to be selected depending on the database connection.
        :param connection_args: {"key": "value"} object with the connection parameters specific for each engine
        :return: created Database object
        """
        ast_query = CreateDatabase(
            name=name,
            engine=engine,
            parameters=connection_args,
        )
        self.api.sql_query(ast_query.to_string())
        return Database(self, name)

    def drop_database(self, name: str):
        """
        Delete integration

        :param name: name of integration
        """
        ast_query = DropDatabase(name=name)
        self.api.sql_query(ast_query.to_string())

    def get_database(self, name: str) -> Database:
        """
        Get integration by name

        :param name: name of integration
        :return: Database object
        """
        if name not in self._list_databases():
            raise AttributeError("Database doesn't exist")
        return Database(self, name)

    def _list_projects(self):
        data = self.api.sql_query("select NAME from information_schema.databases where TYPE='project'")
        return list(data.NAME)

    def list_projects(self) -> List[Project]:
        """
        Show list of project on server

        :return: list of Project objects
        """
        # select * from information_schema.databases where TYPE='project'
        return [Project(self, name) for name in self._list_projects()]

    def create_project(self, name: str) -> Project:
        """
        Create new project and return it

        :param name: name of the project
        :return: Project object
        """

        ast_query = CreateDatabase(
            name=name,
            engine='mindsdb',
            parameters={}
        )

        self.api.sql_query(ast_query.to_string())
        return Project(self, name)

    def drop_project(self, name: str):
        """
        Drop project from server

        :param name: name of the project
        """
        ast_query = DropDatabase(name=name)
        self.api.sql_query(ast_query.to_string())

    def get_project(self, name: str = 'mindsdb') -> Project:
        """
        Get Project by name

        :param name: name of project
        :return: Project object
        """
        if name not in self._list_projects():
            raise AttributeError("Database doesn't exist")
        return Project(self, name)



import pandas as pd

from mindsdb_sql.parser.ast import Select, Star, Identifier, Constant

from mindsdb_sdk.utils import dict_to_binary_op


class Query:
    def __init__(self, api, sql, database=None):
        self.api = api

        self.sql = sql
        self.database = database

    def __repr__(self):
        sql = self.sql.replace('\n', ' ')
        if len(sql) > 40:
            sql = sql[:37] + '...'

        return f'{self.__class__.__name__}({sql})'

    def fetch(self) -> pd.DataFrame:
        """
        Executes query in mindsdb server and returns result
        :return: dataframe with result
        """
        return self.api.sql_query(self.sql, self.database)


class Table(Query):
    def __init__(self, db, name):
        super().__init__(db.api, '', db.name)
        self.name = name
        self.db = db
        self._filters = {}
        self._limit = None
        self._update_query()

    def _filters_repr(self):
        filters = ''
        if len(filters) > 0:
            filters = ', '.join(
                f'{k}={v}'
                for k, v in self._filters
            )
            filters = ', ' + filters
        return filters

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name}{self._filters_repr()})'

    def filter(self, **kwargs):
        """
        Applies filters on table
        table.filter(a=1, b=2) adds where condition to table:
        'select * from table1 where a=1 and b=2'

        :param kwargs: filter
        """
        self._filters.update(kwargs)
        self._update_query()

    def limit(self, val: int):
        """
        Applies limit condition to table query

        :param val: limit size
        """
        self._limit = val
        self._update_query()

    def _update_query(self):
        ast_query = Select(
            targets=[Star()],
            from_table=Identifier(self.name),
            where=dict_to_binary_op(self._filters)
        )
        if self._limit is not None:
            ast_query.limit = Constant(self._limit)
        self.sql = ast_query.to_string()


class View(Table):
    # The same as table
    pass

# TODO getting view sql from api not implemented yet
# class View(Table):
#     def __init__(self, api, data, project):
#         super().__init__(api, data['name'], project)
#         self.view_sql = data['sql']
#
#     def __repr__(self):
#         #
#         sql = self.view_sql.replace('\n', ' ')
#         if len(sql) > 40:
#             sql = sql[:37] + '...'
#
#         return f'{self.__class__.__name__}({self.name}{self._filters_repr()}, sql={sql})'



from functools import wraps

import requests
import pandas as pd


def _try_relogin(fnc):
    wraps(fnc)
    def wrapper(self, *args, **kwargs):
        try:
            return fnc(self, *args, **kwargs)
        except requests.HTTPError as e:
            if e.response.status_code != 401:
                raise e

            # try re-login
            try:
                self.login()
            except requests.HTTPError:
                raise e
            # call once more
            return fnc(self, *args, **kwargs)
    return wrapper


class RestAPI:
    def __init__(self, url=None, email=None, password=None):

        self.url = url
        self.email = email
        self.password = password
        self.session = None

        if email is not None:
            self.login()

    def login(self):
        url = self.url + '/cloud/login'
        json = {'email': self.email, 'password': self.password}
        r = requests.post(url, json=json)
        r.raise_for_status()

        self.session = r.cookies['session']

    @_try_relogin
    def sql_query(self, sql, database=None, lowercase_columns=False):
        if database is None:
            database = 'mindsdb'
        url = self.url + '/api/sql/query'
        r = requests.post(url, json={
            'query': sql,
            'context': {'db': database}
        })
        r.raise_for_status()

        data = r.json()
        if data['type'] == 'table':
            columns = data['column_names']
            if lowercase_columns:
                columns = [i.lower() for i in columns]
            return pd.DataFrame(data['data'], columns=columns)
        if data['type'] == 'error':
            raise RuntimeError(data['error_message'])
        return None

    @_try_relogin
    def projects(self):
        # TODO not used yet

        r = requests.get(self.url + '/api/projects')
        r.raise_for_status()

        return pd.DataFrame(r.json())

    @_try_relogin
    def model_predict(self, project, model, data, params=None, version=None):
        data = data.to_dict('records')

        if version is not None:
            model = f'{model}.{version}'
        if params is None:
            params = {}
        url = self.url + f'/api/projects/{project}/models/{model}/predict'
        r = requests.post(url, json={
            'data': data,
            'params': params
        })
        r.raise_for_status()

        return pd.DataFrame(r.json())

    @_try_relogin
    def objects_tree(self, item=''):
        r = requests.get(self.url + f'/api/tree/{item}')
        r.raise_for_status()

        return pd.DataFrame(r.json())
